#include "gui.h"
#include <string>
#include <OgreKeyEvent.h>
#include <OgreMouseEvent.h>
#include "logic.h"
#include <boost/lexical_cast.hpp>

using namespace CEGUI;
using namespace std;
using namespace boost;

GUI::GUI(Ogre::RenderWindow* window, Ogre::SceneManager* sceneMgr, GameLogic* logic)
: mLogic(logic)
{
  try
  {
    // initialise CEGUI
    mGUIRenderer = new OgreCEGUIRenderer(window, Ogre::RENDER_QUEUE_OVERLAY, false, 3000, sceneMgr);
    mGUISystem = new System(mGUIRenderer);
    Logger::getSingleton().setLoggingLevel(Informative);

    // load GUI data
    SchemeManager::getSingleton().loadScheme("gui.scheme");
    mGUISystem->setDefaultMouseCursor((utf8*)"GUI", (utf8*)"MouseCursor");
    mGUISystem->setDefaultFont((utf8*)"BlueHighway-12");

    // load and setup dialogs
    WindowManager* wMgr = WindowManager::getSingletonPtr();
    mRootWindow = (DefaultWindow*) wMgr->createWindow((utf8*)"DefaultWindow", (utf8*)"RootWindow");
    mGUISystem->setGUISheet(mRootWindow);
    wMgr->loadWindowLayout("MainMenu.layout");
    wMgr->loadWindowLayout("confirm.layout");
    wMgr->loadWindowLayout("highscore.layout");
    wMgr->loadWindowLayout("name.layout");
    // setup button reactions
    PushButton* buttonNew = dynamic_cast<PushButton*>(wMgr->getWindow("Menu/NewGame"));
    buttonNew->subscribeEvent(PushButton::EventClicked, Event::Subscriber(&GUI::onNewGame, this));
    PushButton* buttonHighscore = dynamic_cast<PushButton*>(wMgr->getWindow("Menu/Highscores"));
    buttonHighscore->subscribeEvent(PushButton::EventClicked, Event::Subscriber(&GUI::onHighscore, this));
    PushButton* buttonQuit = dynamic_cast<PushButton*>(wMgr->getWindow("Menu/QuitGame"));
    buttonQuit->subscribeEvent(PushButton::EventClicked, Event::Subscriber(&GUI::onQuit, this));
    PushButton* buttonYes = dynamic_cast<PushButton*>(wMgr->getWindow("Confirm/Yes"));
    buttonYes->subscribeEvent(PushButton::EventClicked, Event::Subscriber(&GUI::onConfirmYes, this));
    PushButton* buttonNo = dynamic_cast<PushButton*>(wMgr->getWindow("Confirm/No"));
    buttonNo->subscribeEvent(PushButton::EventClicked, Event::Subscriber(&GUI::onConfirmNo, this));
    PushButton* buttonClose = dynamic_cast<PushButton*>(wMgr->getWindow("Highscore/Close"));
    buttonClose->subscribeEvent(PushButton::EventClicked, Event::Subscriber(&GUI::onHighscoreClose, this));
    // prepare highscore list
    MultiColumnList* list = dynamic_cast<MultiColumnList*>(wMgr->getWindow("Highscore/List"));
    list->addColumn("RANK", 0, 0.15);
    list->addColumn("NAME", 1, 0.6);
    list->addColumn("SCORE", 2, 0.25);
    // prepare name input box
    Window* editBox = wMgr->getWindow("Name/Name");
    editBox->subscribeEvent(Editbox::EventTextAccepted, Event::Subscriber(&GUI::onNameEntered, this));
  }
  catch (CEGUI::Exception& e)
  {
    OGRE_EXCEPT(Ogre::Exception::ERR_INTERNAL_ERROR, string(e.getMessage().c_str()), "Error parsing menu");
  }
  catch (std::exception& e)
  {
    OGRE_EXCEPT(Ogre::Exception::ERR_INTERNAL_ERROR, string(e.what()), "Error parsing menu");
  }
}

GUI::~GUI()
{
}

void GUI::showMainMenu()
{
  mRootWindow->addChildWindow(WindowManager::getSingleton().getWindow("Menu/Background"));
  MouseCursor::getSingleton().show();
}

void GUI::showConfirmDialog()
{
  mRootWindow->addChildWindow(WindowManager::getSingleton().getWindow("Confirm"));
  MouseCursor::getSingleton().show();
}

void GUI::showHighscore(int pos)
{
  mRootWindow->addChildWindow(WindowManager::getSingleton().getWindow("Highscore"));
  MouseCursor::getSingleton().show();
  MultiColumnList* list = dynamic_cast<MultiColumnList*>(WindowManager::getSingleton().getWindow("Highscore/List"));

  // reenter all highscore entries into the list.
  list->resetList();
  const Highscore& scoreList = mLogic->getHighscore();
  for (size_t i = 0; i < scoreList.numEntries(); ++i)
  {
    ListboxTextItem* rank = new ListboxTextItem(lexical_cast<string>(i+1));
    ListboxTextItem* name = new ListboxTextItem(scoreList.getName(i));
    ListboxTextItem* score = new ListboxTextItem(lexical_cast<string>(scoreList.getScore(i)));
    if (int(i) == pos+1)
    {
      // highlight entry at pos
      rank->setTextColours(colour(1.0, 0.75, 0.37));
      name->setTextColours(colour(1.0, 0.75, 0.37));
      score->setTextColours(colour(1.0, 0.75, 0.37));
    }
    unsigned int row = list->addRow(rank, 0);
    list->setItem(name, 1, row);
    list->setItem(score, 2, row);
  }
}

void GUI::showNameDialog(int score)
{
  mRootWindow->addChildWindow(WindowManager::getSingleton().getWindow("Name"));
  MouseCursor::getSingleton().show();
  // clear previous entries
  Editbox* nameBox = dynamic_cast<Editbox*>(WindowManager::getSingleton().getWindow("Name/Name"));
  nameBox->setText("");
  nameBox->activate();
  // save score here to submit it to the highscore list later.
  mScore = score;
}


bool GUI::onNewGame(const EventArgs& e)
{
  mRootWindow->removeChildWindow(WindowManager::getSingleton().getWindow("Menu/Background"));
  MouseCursor::getSingleton().hide();
  mLogic->startGame();
  return true;
}

bool GUI::onHighscore(const EventArgs& e)
{
  mRootWindow->removeChildWindow(WindowManager::getSingleton().getWindow("Menu/Background"));
  showHighscore(-1);
  return true;
}

bool GUI::onQuit(const EventArgs& e)
{
  mLogic->quit();
  return true;
}

bool GUI::onConfirmYes(const EventArgs& e)
{
  // user aborted game.
  mRootWindow->removeChildWindow(WindowManager::getSingleton().getWindow("Confirm"));
  mLogic->endGame(false);
  return true;
}

bool GUI::onConfirmNo(const EventArgs& e)
{
  mRootWindow->removeChildWindow(WindowManager::getSingleton().getWindow("Confirm"));
  MouseCursor::getSingleton().hide();
  mLogic->continueGame();
  return true;
}

bool GUI::onHighscoreClose(const EventArgs& e)
{
  mRootWindow->removeChildWindow(WindowManager::getSingleton().getWindow("Highscore"));
  mRootWindow->addChildWindow(WindowManager::getSingleton().getWindow("Menu/Background"));
  return true;
}

bool GUI::onNameEntered(const EventArgs& e)
{
  Editbox* nameBox = dynamic_cast<Editbox*>(WindowManager::getSingleton().getWindow("Name/Name"));
  string name = nameBox->getText().c_str();
  // insert highscore
  int pos = mLogic->getHighscore().add(mScore, name);
  mRootWindow->removeChildWindow(WindowManager::getSingleton().getWindow("Name"));
  // show highscore list
  showHighscore(pos);
  return true;
}


MouseButton GUI::ConvertOgreButtonToCegui(int buttonID)
{
  switch (buttonID)
  {
    case Ogre::MouseEvent::BUTTON0_MASK:
      return LeftButton;
    case Ogre::MouseEvent::BUTTON1_MASK:
      return RightButton;
    case Ogre::MouseEvent::BUTTON2_MASK:
      return MiddleButton;
    case Ogre::MouseEvent::BUTTON3_MASK:
      return X1Button;
    default:
      return LeftButton;
  }
}

void GUI::keyPressed(Ogre::KeyEvent* e)
{
  mGUISystem->injectKeyDown(e->getKey());
  mGUISystem->injectChar(e->getKeyChar());
  e->consume();
}

void GUI::keyReleased(Ogre::KeyEvent* e)
{
  mGUISystem->injectKeyUp(e->getKey());
  e->consume();
}

void GUI::mouseMoved(Ogre::MouseEvent* e)
{
  mGUISystem->injectMouseMove(e->getRelX()*mGUIRenderer->getWidth(),
    e->getRelY()*mGUIRenderer->getHeight());
  e->consume();
}

void GUI::mousePressed(Ogre::MouseEvent* e)
{
  mGUISystem->injectMouseButtonDown(ConvertOgreButtonToCegui(e->getButtonID()));
  e->consume();
}

void GUI::mouseReleased(Ogre::MouseEvent* e)
{
  mGUISystem->injectMouseButtonUp(ConvertOgreButtonToCegui(e->getButtonID()));
  e->consume();
}

void GUI::mouseDragged(Ogre::MouseEvent* e)
{
  mouseMoved(e);
}
