#include "logic.h"
#include "player.h"
#include "mouse.h"
#include "needle.h"
#include "score.h"
#include "object.h"
#include "gui.h"
#include "combos.h"
#include <OgreFont.h>
#include <OgreFontManager.h>
#include <boost/lexical_cast.hpp>
#include <boost/format.hpp>
#include "OgreAudiere.h"

using namespace boost;

const double SPAWNTIME_MIN = 1.5;
const double SPAWNTIME_MAX = 3.8;
const double THROW_ACCEL = 20.0;
const double THROW_MAX = 50.0;
const double THROW_MIN = 10.0;
const double GAME_TIME = 120.0;
const int NUM_COLOURS = 5;
const size_t HIT_HISTORY = 5;
const int NUM_SOUNDS_MUSIC = 1;
const int NUM_SOUNDS_CAT = 3;

GameLogic::GameLogic(RenderWindow* win, Camera* cam, SceneManager* sceneMgr)
: ExampleFrameListener(win, cam, true, true), mRunning(false), mStarted(false), mQuit(false),
  mPlayer(0), mSceneMgr(sceneMgr), mPrepareThrow(false), mScoreDisplay(0), mHitsChain(1),
  mShowNotice(0.0), mHighscore("highscores.dat")
{
  mEventProcessor->addMouseListener(this);
  mEventProcessor->addMouseMotionListener(this);

  // create GUI object
  mGUI = new GUI(win, sceneMgr, this);
  mEventProcessor->addKeyListener(mGUI);
  mEventProcessor->addMouseListener(mGUI);
  mEventProcessor->addMouseMotionListener(mGUI);

  // load combinations
  mCombos = new Combinations;

  // initialise sound system
  mAudioDevice = audiere::OpenDevice();
  mBkgMusic = audiere::OpenSound(mAudioDevice, audiere::OpenOgreResourceFile("music.ogg"), true);
  mSoundCombo = audiere::OpenSound(mAudioDevice, audiere::OpenOgreResourceFile("combo.wav"));
}

GameLogic::~GameLogic()
{
  delete mPlayer;
  delete mScoreDisplay;
  delete mGUI;
  delete mCombos;
  ObjectManager::instance().deleteAll();
}

void GameLogic::startGame()
{

  // create player mouse
  mPlayer = new Player(mSceneMgr, mAudioDevice);

  // initialise the mouse spawn list
  addToSpawnList(5);
  updateSpawnList(0.0);

  // reset initial score
  mScoreDisplay = new MovableText("Score_Display", "0", "BlueHighway", 4, ColourValue(0x99/255.0, 0xc9/255.0, 0xe0/255.0));
  mSceneMgr->getEntity("Cat1")->attachObjectToBone("Belly", mScoreDisplay, Quaternion::IDENTITY, Vector3(1.5, 15.0, 0.0));
  mScore = 0;
  updateScore(0);

  selectCatAnimation();

  mGameTime = GAME_TIME;
  initHUDElements();
  updateClockDisplay();
  updateHitDisplay();

  mBkgMusic->reset();
  mBkgMusic->play();

  mStarted = true;
  mRunning = true;
}


void GameLogic::pauseGame()
{
  resetHUDElements();
  mPlayer->hide();
  ObjectManager::instance().hideAll();
  mBkgMusic->stop();
  mRunning = false;
}

void GameLogic::continueGame()
{
  initHUDElements();
  mPlayer->show();
  ObjectManager::instance().showAll();
  mBkgMusic->play();
  mRunning = true;
}

void GameLogic::endGame(bool highscore)
{
  // remove all game objects
  mBkgMusic->stop();
  mSpawnList.clear();
  mLastHits.clear();
  delete mPlayer;
  mPlayer = 0;
  delete mScoreDisplay;
  mScoreDisplay = 0;
  ObjectManager::instance().deleteAll();
  resetHUDElements();
  mStarted = false;
  mRunning = false;
  if (highscore)
    mGUI->showNameDialog(mScore);
  else
    mGUI->showMainMenu();
}



void GameLogic::initHUDElements()
{
  OverlayManager* overlayMgr = OverlayManager::getSingletonPtr();
  overlayMgr->getByName("UpcomingOverlay")->show();
  overlayMgr->getByName("ForceOverlay")->show();
  overlayMgr->getByName("LastHitsOverlay")->show();
  overlayMgr->getByName("ClockOverlay")->show();
  overlayMgr->getByName("NotifyOverlay")->show();
}

void GameLogic::resetHUDElements()
{
  OverlayManager* overlayMgr = OverlayManager::getSingletonPtr();
  overlayMgr->getByName("UpcomingOverlay")->hide();
  overlayMgr->getByName("ForceOverlay")->hide();
  overlayMgr->getByName("LastHitsOverlay")->hide();
  overlayMgr->getByName("ClockOverlay")->hide();
  overlayMgr->getByName("NotifyOverlay")->hide();
}


void GameLogic::updateClockDisplay()
{
  TextAreaOverlayElement* clock = dynamic_cast<TextAreaOverlayElement*>
    (OverlayManager::getSingleton().getOverlayElement("Clock"));
  clock->setCaption((format("%d:%02d") % (int(mGameTime)/60) % (int(mGameTime)%60)).str());
}


void GameLogic::addToSpawnList(int n)
{
  for (int i = 0; i < n; ++i)
  {
    SpawnInfo info;
    info.countdown = Math::RangeRandom(SPAWNTIME_MIN, SPAWNTIME_MAX);
    info.colour = int(Math::RangeRandom(0, NUM_COLOURS));
    mSpawnList.push_back(info);
  }
}

void GameLogic::updateSpawnList(double passedTime)
{
  if (!mSpawnList.empty())
  {
    SpawnInfo& info = *mSpawnList.begin();
    info.countdown -= passedTime;
    if (info.countdown <= 0.0)
    {
      ObjectManager::instance().add(new Mouse(mSceneMgr, this, info.colour, mAudioDevice));
      mSpawnList.erase(mSpawnList.begin());
      addToSpawnList(1);
    }

    for (size_t i = 0; i < mSpawnList.size(); ++i)
    {
      OverlayElement* elem = OverlayManager::getSingleton().getOverlayElement(
        "Upcoming/Ballon."+lexical_cast<String>(i));
      elem->setMaterialName("HUD/Ballon."+lexical_cast<String>(mSpawnList[i].colour));
    }
  }
}


void GameLogic::quit()
{
  mQuit = true;
}

void GameLogic::showMainMenu()
{
  mGUI->showMainMenu();
}

void GameLogic::updateScore(int addScore)
{
  mScore += addScore;
  mScoreDisplay->setCaption(lexical_cast<String>(mScore));
}

void GameLogic::notifyHit(int colour)
{
  // store the hit in our hit history and check for triggered combos.
  mLastHits += lexical_cast<String>(colour);
  if (mLastHits.size() > HIT_HISTORY)
  {
    mLastHits.erase(0, mHitsChain);
    mHitsChain = 1;
  }
  for (size_t i = 0; i < mLastHits.size(); ++i)
  {
    if (i != 0 && i < mHitsChain)
      continue;
    String check = mLastHits.substr(i, mLastHits.size());
    Combinations::Info* info = mCombos->getCombo(check);
    if (info)
    {
      // triggered valid combination
      mSoundCombo->play();
      mLastHits.erase(0, i);
      mHitsChain = check.size();
      updateScore(info->score);
      String caption = info->name + " (+" + lexical_cast<String>(info->score) + ")";
      OverlayManager::getSingleton().getOverlayElement("Notify")->setCaption(caption);
      mShowNotice = 3.0;
    }
  }

  updateHitDisplay();
}

void GameLogic::updateHitDisplay()
{
  // update last hit display
  for (size_t i = 0; i < HIT_HISTORY; ++i)
  {
    int pos = mLastHits.size() - i - 1;
    if (pos < 0)
      OverlayManager::getSingleton().getOverlayElement("LastHits/Ballon."+lexical_cast<String>(i))
        ->setMaterialName("HUD/Invisible");
    else
    {
      String ballon = "HUD/Ballon.";
      if (pos < mHitsChain && mHitsChain > 1)
        ballon += "X.";
      ballon += mLastHits[pos];
      OverlayManager::getSingleton().getOverlayElement("LastHits/Ballon."+lexical_cast<String>(i))
        ->setMaterialName(ballon);
    }
  }
}

void GameLogic::keyPressed(KeyEvent* e)
{
  if (mStarted)
  {
    switch (e->getKey())
    {
      case KC_ESCAPE:
        if (mRunning)
        {
          pauseGame();
          mGUI->showConfirmDialog();
        }
        break;

      case KC_LEFT:
        mPlayer->moveLeft(true);
        break;

      case KC_RIGHT:
        mPlayer->moveRight(true);
        break;

      case KC_SPACE:
        mPrepareThrow = true;
        mThrowSpeed = THROW_MIN;
        break;

      default: break; // ignore
    }
    e->consume();
  }
}


void GameLogic::selectCatAnimation()
{
  // randomly selects 1 of 3 animations for the cat.
  int anim = int(Math::RangeRandom(1.0, 4.0));
  mCatAnim = mSceneMgr->getEntity("Cat1")->getAnimationState("Idle"+lexical_cast<String>(anim));
  mCatAnim->setEnabled(true);
  mCatAnim->setLoop(false);
  mCatAnim->setTimePosition(0.0);
  // play a random cat sound
  int catSound = int(Math::RangeRandom(0, NUM_SOUNDS_CAT));
  mSoundCat = audiere::OpenSound(mAudioDevice, audiere::OpenOgreResourceFile(str(format("cat.%s.wav") % catSound)));
  mSoundCat->play();
}

void GameLogic::updateCatAnimation(double passedTime)
{
  mCatAnim->addTime(passedTime);
  if (Math::RealEqual(mCatAnim->getTimePosition(), mCatAnim->getLength()))
    selectCatAnimation();
}


void GameLogic::keyReleased(KeyEvent* e)
{
  if (mStarted)
  {
    switch (e->getKey())
    {
      case KC_LEFT:
        mPlayer->moveLeft(false);
        break;

      case KC_RIGHT:
        mPlayer->moveRight(false);
        break;

      case KC_SPACE:
        mPrepareThrow = false;
        mPlayer->throwNeedle(mThrowSpeed);
        OverlayManager::getSingleton().getOverlayElement("Force/Bar")->setWidth(0.0);
        break;

      default: break; // ignore
    }
    e->consume();
  }
}


bool GameLogic::frameStarted(const FrameEvent& evt)
{
  // update Audiere
  mAudioDevice->update();

  if (mRunning)
  {
    mGameTime -= evt.timeSinceLastFrame;
    if (mGameTime <= 0.0)
    {
      endGame(true);
      return true;
    }
    updateClockDisplay();

    // update player object state
    mPlayer->update(evt.timeSinceLastFrame);

    // countdown for next mouse creation
    updateSpawnList(evt.timeSinceLastFrame);

    // update objects state
    ObjectManager::instance().update(evt.timeSinceLastFrame);

    // preparing to throw?
    if (mPrepareThrow)
    {
      mThrowSpeed += THROW_ACCEL * evt.timeSinceLastFrame;
      if (mThrowSpeed > THROW_MAX)
        mThrowSpeed = THROW_MAX;
      // update graphical representation
      OverlayElement* bar = OverlayManager::getSingleton().getOverlayElement("Force/Bar");
      bar->setWidth(0.5 * (mThrowSpeed - THROW_MIN) / (THROW_MAX - THROW_MIN));
      bar->setLeft(-bar->getWidth()/2);
    }

    // update notice display
    if (mShowNotice > 0.0)
    {
      mShowNotice -= evt.timeSinceLastFrame;
      if (mShowNotice <= 0.0)
        OverlayManager::getSingleton().getOverlayElement("Notify")->setCaption("");
    }

    updateCatAnimation(evt.timeSinceLastFrame);
  }

  // continue rendering?
  return !mQuit;
}


Highscore& GameLogic::getHighscore()
{
  return mHighscore;
}
