#include "mouse.h"
#include "constants.h"
#include "score.h"
#include <boost/lexical_cast.hpp>
#include <boost/format.hpp>
#include <OgreTagPoint.h>
#include "OgreAudiere.h"

using namespace Ogre;
using namespace boost;
using namespace std;

const int NUM_TEXTURES = 3;
const int NUM_SOUNDS_BLOW = 1;
const int NUM_SOUNDS_ESCAPE = 1;
const int NUM_SOUNDS_HIT = 3;
const int NUM_SOUNDS_CRASHED = 2;

unsigned int numberOfMice = 0;

Mouse::Mouse(SceneManager* sceneMgr, GameLogic* logic, int colour, audiere::AudioDevicePtr audioDev)
: mSceneMgr(sceneMgr), mLogic(logic), mNode(0), mStartNode(0), mEnt(0), mBallon(0),
  mAccel(0.0), mSpeed(0.0), mIsDead(false), mColour(colour)
{
  // choose spawn point
  int spawn = int(Math::RangeRandom(0, NUM_SPAWNS));
  mSpawnNode = mSceneMgr->getSceneNode("Spawn."+lexical_cast<string>(spawn));
  mNode = mSceneMgr->getRootSceneNode()->createChildSceneNode();
  mNode->setPosition(mSpawnNode->getPosition());

  // choose starting point for flying
  SceneNode* startNode = mSceneMgr->getSceneNode("Start_Area");
  mStartNode = mSceneMgr->getRootSceneNode()->createChildSceneNode();
  mStartNode->setPosition(startNode->getPosition());
  mStartNode->translate(Vector3::UNIT_X * Math::RangeRandom(X_MIN, X_MAX));
  mStartNode->translate(Vector3::UNIT_Z * Math::RangeRandom(Z_MIN, Z_MAX));

  // create mouse entity
  mEnt = mSceneMgr->createEntity("Mouse."+lexical_cast<string>(numberOfMice++), "Flying_Mouse.mesh");
  mEnt->setQueryFlags(MOUSE_FLAG);
  mEnt->setUserObject(this);
  mNode->attachObject(mEnt);
  int texture = int(Math::RangeRandom(0, NUM_TEXTURES));
  mEnt->setMaterialName("Mouse."+lexical_cast<String>(texture));
  // rotate towards starting point
  Vector3 src = mNode->getOrientation() * Vector3::UNIT_X;
  mNode->rotate(Quaternion (src.getRotationTo(mStartNode->getPosition() - mNode->getPosition())));
  // set walking animation
  mAnim = mEnt->getAnimationState("Walk");
  mAnim->setLoop(true);
  mAnim->setEnabled(true);
  
  // load sounds
  using namespace audiere;
  // choose randomly from available sounds.
  int blow = int(Math::RangeRandom(0, NUM_SOUNDS_BLOW));
  int escape = int(Math::RangeRandom(0, NUM_SOUNDS_ESCAPE));
  int hit = int(Math::RangeRandom(0, NUM_SOUNDS_HIT));
  int crashed = int(Math::RangeRandom(0, NUM_SOUNDS_CRASHED));
  mSoundBlow = OpenSound(audioDev, OpenOgreResourceFile(str(format("blow.%s.wav") % blow)));
  mSoundHit = OpenSound(audioDev, OpenOgreResourceFile(str(format("hit.%s.wav") % hit)));
  mSoundCrashed = OpenSound(audioDev, OpenOgreResourceFile(str(format("crashed.%s.wav") % crashed)));
  mSoundEscaped = OpenSound(audioDev, OpenOgreResourceFile(str(format("escape.%s.wav") % escape)));

  mState = STATE_MOVING;
}


Mouse::~Mouse()
{
  if (mNode)
    mSceneMgr->destroySceneNode(mNode->getName());
  if (mStartNode)
    mSceneMgr->destroySceneNode(mStartNode->getName());
  if (mEnt)
    mSceneMgr->destroyEntity(mEnt);
  if (mBallon)
    mSceneMgr->destroyEntity(mBallon);
}


void Mouse::update(double passedTime)
{
  mAnim->addTime(passedTime);

  switch (mState)
  {
    case STATE_MOVING:
      stateMoving(passedTime);
      break;
    case STATE_PREPARING:
      statePreparing(passedTime);
      break;
    case STATE_FLYING:
      stateAir(passedTime, true);
      break;
    case STATE_FALLING:
      stateAir(passedTime, false);
      break;
    case STATE_CRASHED:
      stateCrashed(passedTime);
      break;
    case STATE_WALKAWAY:
      stateWalkAway(passedTime);
    default: break;
  }
}

void Mouse::stateMoving(double passedTime)
{
  // move towards starting point.
  Vector3 direction = mStartNode->getPosition() - mNode->getPosition();
  double distance = direction.normalise();
  double move = WALK_SPEED * passedTime;
  if (distance - move <= 0.0)
  {
    // reached destination
    mNode->setPosition(mStartNode->getPosition());
    mNode->setOrientation(Quaternion(Radian(0), Vector3::UNIT_X));
    mAnim->setEnabled(false);
    mAnim = mEnt->getAnimationState("Blow");
    mAnim->setLoop(false);
    mAnim->setEnabled(true);
    // choose a ballon size etc.
    mBallonSize = Math::RangeRandom(1.0, 2.0);
    // create ballon and attach to bone.
    mBallon = mSceneMgr->createEntity("Ballon."+mEnt->getName(), "Ballon.mesh");
    mBallon->setMaterialName("Ballon."+lexical_cast<string>(mColour));
    mBallon->setQueryFlags(BALLON_FLAG);
    mBallon->setUserObject(this);
    TagPoint* tp = mEnt->attachObjectToBone("Ballon", mBallon);
    tp->setScale(mBallonSize, mBallonSize, mBallonSize);
    AnimationState* anim = mBallon->getAnimationState("Expand");
    anim->setLoop(false);
    anim->setEnabled(true);
    // play blow sound
    mSoundBlow->play();

    mState = STATE_PREPARING;
  } else
  {
    mNode->translate(direction * move);
  }
}

void Mouse::statePreparing(double passedTime)
{
  mBallon->getAnimationState("Expand")->addTime(passedTime);
  // wait until preparation animation has ended.
  if (Math::RealEqual(mAnim->getTimePosition(), mAnim->getLength()))
  {
    // load floating animation.
    mAnim->setEnabled(false);
    mAnim = mEnt->getAnimationState("Floating");
    mAnim->setLoop(true);
    mAnim->setEnabled(true);
    mAnim->setTimePosition(0);
    // set acceleration and speed
    mAccel = BASE_FLOATING_ACCEL * mBallonSize * mBallonSize;
    mSpeed = 0.0;

    mState = STATE_FLYING;
  }
}

void Mouse::stateAir(double passedTime, bool flying)
{
  mSpeed += mAccel * passedTime;

  if (Math::RealEqual(mAnim->getTimePosition(), mAnim->getLength()) && !mAnim->getLoop())
  {
    mAnim->setEnabled(false);
    mAnim = mEnt->getAnimationState("Falling");
    mAnim->setLoop(true);
    mAnim->setEnabled(true);
  }

  mNode->translate(Vector3::UNIT_Y * mSpeed * passedTime);

  if (flying && mNode->getPosition().y >= 100.0)
  {
    // mouse escaped
    if (!mIsDead)
      mSoundEscaped->play();
    mIsDead = true;
  }

  if (!flying && mNode->getPosition().y <= 0.0)
  {
    // hit the ground
    mNode->setPosition(Vector3(mNode->getPosition().x, 0.0, mNode->getPosition().z));
    mAnim->setEnabled(false);
    mAnim = mEnt->getAnimationState("Hit_Ground");
    mAnim->setEnabled(true);
    mAnim->setLoop(false);
    // play crashed sound
    mSoundCrashed->play();
    // create stars
    mStars = mSceneMgr->createEntity("Stars."+mEnt->getName(), "Stars.mesh");
    mStars->setCastShadows(false);
    mEnt->attachObjectToBone("Head", mStars);
    AnimationState* anim = mStars->getAnimationState("Rotate");
    anim->setEnabled(true);
    anim->setLoop(false);
    mState = STATE_CRASHED;
  }
}

void Mouse::stateCrashed(double passedTime)
{
  mStars->getAnimationState("Rotate")->addTime(passedTime);
  // wait until animation is over.
  if (Math::RealEqual(mAnim->getTimePosition(), mAnim->getLength()))
  {
    mAnim->setEnabled(false);
    mAnim = mEnt->getAnimationState("Walk");
    mAnim->setLoop(true);
    mAnim->setEnabled(true);
    // rotate towards spawn point
    Vector3 src = mNode->getOrientation() * Vector3::UNIT_X;
    mNode->rotate(Quaternion (src.getRotationTo(mSpawnNode->getPosition() - mNode->getPosition())));
    mSceneMgr->destroyEntity(mStars);
    mState = STATE_WALKAWAY;
  }
}

void Mouse::stateWalkAway(double passedTime)
{
  // return to spawn point
  Vector3 direction = mSpawnNode->getPosition() - mNode->getPosition();
  double distance = direction.normalise();
  double move = WALK_SPEED * passedTime;
  if (distance - move <= 0.0)
  {
    // reached destination
    mIsDead = true;
  } else
  {
    mNode->translate(direction * move);
  }
}

bool Mouse::isDead() const
{
  return mIsDead && !mSoundEscaped->isPlaying();
}

void Mouse::show()
{
  mEnt->setVisible(true);
  if (mBallon)
    mBallon->setVisible(true);
}

void Mouse::hide()
{
  mEnt->setVisible(false);
  if (mBallon)
    mBallon->setVisible(false);
}

void Mouse::killBallon()
{
  if (mBallon && mState == STATE_FLYING)
  {
    mAnim->setEnabled(false);
    mAnim = mEnt->getAnimationState("Start_Falling");
    mAnim->setLoop(false);
    mAnim->setEnabled(true);
    mAccel = GRAVITY;
    mState = STATE_FALLING;
    mSceneMgr->destroyEntity(mBallon);
    mBallon = 0;
    // play hit sound
    mSoundHit->play();

    // insert score display
    SceneNode* scoreNode = mSceneMgr->getRootSceneNode()->createChildSceneNode();
    scoreNode->setPosition(mNode->getPosition());
    ObjectManager::instance().add(new Score(mSceneMgr, scoreNode, "+10"));
    // update game score
    mLogic->updateScore(10);
    mLogic->notifyHit(mColour);
  }
}
