#include "player.h"
#include "needle.h"
#include "constants.h"
#include <boost/lexical_cast.hpp>

using namespace Ogre;
using namespace boost;
using namespace std;

unsigned int numNeedle = 0;

Player::Player(SceneManager* sceneMgr, audiere::AudioDevicePtr audioDev)
: mSceneMgr(sceneMgr), mMoveLeft(false), mMoveRight(false),
  mPosition(0.0), mThrowing(false), mAudioDevice(audioDev)
{
  // create player mouse object.
  mObj = mSceneMgr->createEntity("Player", "Mouse.mesh");
  mNode = mSceneMgr->getSceneNode("Player_Plane")->createChildSceneNode();
  mNode->attachObject(mObj);
  // give it a needle.
  mNeedle = mSceneMgr->createEntity("Needle."+lexical_cast<string>(numNeedle++), "Needle.mesh");
  mObj->attachObjectToBone("Needle", mNeedle);
  // set to idle animation.
  mAnim = mObj->getAnimationState("Idle");
  mAnim->setLoop(true);
  mAnim->setEnabled(true);

  mThrowAnim = mObj->getAnimationState("Throw");
}

Player::~Player()
{
  mSceneMgr->destroyEntity(mObj);
}

void Player::moveLeft(bool move)
{
  mMoveLeft = move;
  setMoveAnimation();
}

void Player::moveRight(bool move)
{
  mMoveRight = move;
  setMoveAnimation();
}

void Player::setMoveAnimation()
{
  string animName;
  if (mMoveLeft == mMoveRight)
  {
    // load idle animation
    animName = "Idle";
  }
  else if (mMoveLeft)
  {
    // load walk left animation
    animName = "Step_L";
  }
  else
  {
    // load walk right animation
    animName = "Step_R";
  }
  mAnim->setEnabled(false);
  mAnim = mObj->getAnimationState(animName);
  mAnim->setLoop(true);
  mAnim->setEnabled(true);
}

void Player::update(double passedTime)
{
  mAnim->addTime(passedTime);

  if (mThrowing)
  {
    mThrowAnim->addTime(passedTime);
    if (Math::RealEqual(mThrowAnim->getTimePosition(), mThrowAnim->getLength()))
    {
      LogManager::getSingleton().logMessage("[PLAYER] Finished throwing animation.");
      // create scene node for needle.
      SceneNode* node = mSceneMgr->getRootSceneNode()->createChildSceneNode();
      Node* needleParent = mNeedle->getParentNode();
      node->setPosition(needleParent->getWorldPosition());
      node->setOrientation(needleParent->getOrientation());

      // detach needle
      LogManager::getSingleton().logMessage("[PLAYER] Detaching needle "+mNeedle->getName()+" from player mesh.");
      mObj->detachObjectFromBone(mNeedle);
      ObjectManager::instance().add(new Needle(mSceneMgr, node, mNeedle, mThrowSpeed, mAudioDevice));
      // create new needle
      mNeedle = mSceneMgr->createEntity("Needle."+lexical_cast<string>(numNeedle++), "Needle.mesh");
      LogManager::getSingleton().logMessage("[PLAYER] Created new needle "+mNeedle->getName());
      mObj->attachObjectToBone("Needle", mNeedle);

      mThrowing = false;
      setMoveAnimation();
    }
  }

  // move
  if (mMoveLeft)
  {
    mPosition -= MOVE_SPEED * passedTime;
    if (mPosition < WAY_MIN)
      mPosition = WAY_MIN;
    mNode->setPosition(Vector3(0.0, 0.0, -mPosition));
  }
  if (mMoveRight)
  {
    mPosition += MOVE_SPEED * passedTime;
    if (mPosition > WAY_MAX)
      mPosition = WAY_MAX;
    mNode->setPosition(Vector3(0.0, 0.0, -mPosition));
  }
}

void Player::throwNeedle(double speed)
{
  if (!mThrowing)
  {
    mThrowSpeed = speed;
    mThrowing = true;
    // start throwing animation
    mThrowAnim->setTimePosition(0.0);
    mThrowAnim->setLoop(false);
    mThrowAnim->setEnabled(true);
  }
}
